"""
This module hides the source of implementation of weak ref to a method: for Python 3.4, it is
Python's weakref module; for earlier Python, it is the weakrefmethod module from PyPI.
Prior to pypubsub 4.0, WeakMethod was a custom class that adhered to the WeakRef API.

Use the getWeakRef(object) module function to create the
proper type of weak reference (weakref.WeakRef or WeakMethod) for given object.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from inspect import ismethod
from weakref import ref as WeakRef

# for weakly bound methods:
try:
    from weakref import WeakMethod
except:
    from weakrefmethod import WeakMethod

# type hinting:
from typing import Tuple, List, Sequence, Mapping, Dict, Callable, Any, Optional, Union, TextIO


WeakObjOrMethod = Union[WeakMethod, WeakRef]

DeadRefObserver = Callable[[WeakObjOrMethod], None]


def getWeakRef(obj, notifyDead: DeadRefObserver = None):
    """
    Get a weak reference to obj. If obj is a bound method, a WeakMethod
    object, that behaves like a WeakRef, is returned; if it is
    anything else a WeakRef is returned. If obj is an unbound method,
    a ValueError will be raised.
    """
    if ismethod(obj):
        createRef = WeakMethod
    else:
        createRef = WeakRef

    return createRef(obj, notifyDead)



"""
Skeleton code (will not load... must be completed):
UDP-based net-centric Pubsub extension so that pubsub can be used over a
network of pubsub-powered applications. Extending to use TCP wouldn't be
too difficult.

Usage: each of your applications that uses pubsub would create a
suitable AppListener, then call its sendMessage when wanting to
send a message, or call processMessages whenever it is ready to
process queued messages from other applications. In addition, one
of your applications must be the "pubsub server", or the server
could live in its own python program.  For it, you would
create a PubsubAppServer then call its processMessages()
in an infinite loop.

Oliver Schoenborn
"""

from socket import socket, timeout
from pubsub import pub


class Marshaller:
    ...
    depends
    on
    your
    strategy / method...
    ...
    for instance, could use unpickle...
    ... or XML, SOAP, your
    own
    format, etc...

    def pack(self, topic, **args):
        ...
        pack
        topic and args
        into
        a
        data
        string...
        return data

    def unpack(self, data):
        ...
        unpack
        data
        into
        a
        topic and args
        return topic, msgData

    def getTopic(self, data):
        ...
        get
        topic
        from data...
        return topic


class AppListener:
    def __init__(self, server, readPort, maxWait=0.01):
        self.__server = server
        self.__marshaller = Marshaller()
        self.__udpSocket = socket(...
        server...)
        self.__udpSocket.bind(...
        readPort...)
        self.__udpSocket.settimeout(maxWait)  # want limited blocking

    def subscribe(self, topic):
        self.__udpSocket.write('subscribe %s' % topic)

    def processMessages(self):
        bufsize = 4096
        more = True
        while more:
            try:
                data, sender = self.__udpSocket.recvfrom(bufsize)
                if sender == self.__server:
                    self.__publish(data)
            except timeout:
                more = False

    def __publish(self, data):
        topic, argsValues = self.__marshaller.unpack(data)
        pub.sendMessage(topic, **argsValues)

    def sendMessage(topic, **msgData):
        packedData = self.__marshaller.pack(msgData)
        self.server.write(packedData)


class PubsubAppServer:
    SERVER_PORT = 8743  # pick to suit you

    def __init__(self):
        self.__listenerApps = {}  # will be list of sets by topic
        self.__marshaller = Marshaller()
        self.__udpSocket = socket(...)
        self.__udpSocket.bind(...
        SERVER_PORT...)

        def processMessages(self):
            bufsize = 4096
            more = True
            while more:
                try:
                    data, sender = self.__udpSocket.recv(bufsize)
                    if data.startswith('subscribe'):
                        self.__register(sender, data)
                    else:
                        self.__dispatch(data)
                except timeout:
                    more = False

        def __register(self, sender, data):
            topic = data.split(' ')[1]  # second item
            self.__listenerApps.setdefault(topic, set()).add(sender)

        def __dispatch(self, data):
            topic = self.__marshaller.getTopic(data)
            # send to all registered apps for that topic;
            # if no such topic then do nothing:
            for listenerApp in self.__listenerApps.get(topic, []):
                self.__udpSocket.write(data)



"""
Pubsub package initialization.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

# Version is provided by setuptools_scm at build time via src/pubsub/_version.py
try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover - fallback for editable installs without SCM metadata
    __version__ = "0.0.0"

__all__ = [
    'pub',
    'utils',
    '__version__'
]



"""
Definitions related to message data specification.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

import weakref
from typing import Tuple, List, Sequence as Seq, Mapping, Dict, Callable, Any, Optional, Union

from .topicutils import stringize, WeakNone
from .annotations import annotationType
from .topicexc import MessageDataSpecError
from .listener import getArgs as getListenerArgs, UserListener


ArgsDocs = Dict[str, str]
MsgData = Mapping[str, Any]


def verifyArgsDifferent(allArgs, allParentArgs, topicName):
    """
    Verify that allArgs does not contain any of allParentArgs. Raise
    MessageDataSpecError if fail.
    """
    extra = set(allArgs).intersection(allParentArgs)
    if extra:
        msg = 'Args %%s already used in parent of "%s"' % topicName
        raise MessageDataSpecError(msg, tuple(extra))


def verifySubset(all, sub, topicName, extraMsg=''):
    """
    Verify that sub is a subset of all for topicName. Raise
    MessageDataSpecError if fail.
    """
    notInAll = set(sub).difference(all)
    if notInAll:
        args = ','.join(all)
        msg = 'Params [%s] missing inherited [%%s] for topic "%s"%s' % (args, topicName, extraMsg)
        raise MessageDataSpecError(msg, tuple(notInAll))


def topicArgsFromCallable(_callable: UserListener, ignoreArgs: Seq[str] = ()) -> Tuple[ArgsDocs, List[str]]:
    """
    Get the topic message data names and list of those that are required,
    by introspecting given callable. Returns a pair, (args, required)
    where args is a dictionary of allowed message data names vs docstring,
    and required states which ones are required rather than optional.
    """
    argsInfo = getListenerArgs(_callable, ignoreArgs=ignoreArgs)
    required = argsInfo.getRequiredArgs()
    defaultDoc = 'UNDOCUMENTED'
    args = dict.fromkeys(argsInfo.allParams, defaultDoc)
    return args, required


class ArgSpecGiven:
    """
    The message data specification (MDS) for a topic.

    This consists of each argument name that listener should have in its
    signature, plus which ones are required in any sendMessage(), and a
    documentation string for each argument. This instance will be transformed
    into an ArgsInfo object which is basically a superset of that information,
    needed to ensure that the arguments specifications satisfy
    pubsub policies for chosen API version.
    """

    SPEC_GIVEN_NONE = 1  # specification not given
    SPEC_GIVEN_ALL = 3  # all args specified

    def __init__(self, argsDocs: ArgsDocs = None, reqdArgs: Seq[str] = None):
        self.reqdArgs = tuple(reqdArgs or ())

        if argsDocs is None:
            self.argsSpecType = ArgSpecGiven.SPEC_GIVEN_NONE
            self.argsDocs = {}
        else:
            self.argsSpecType = ArgSpecGiven.SPEC_GIVEN_ALL
            self.argsDocs = argsDocs

            # check that all args marked as required are in argsDocs
            missingArgs = set(self.reqdArgs).difference(self.argsDocs.keys())  # py3: iter keys ok
            if missingArgs:
                msg = 'Params [%s] missing inherited required args [%%s]' % ','.join(argsDocs.keys())  # iter keys ok
                raise MessageDataSpecError(msg, missingArgs)

    def setAll(self, allArgsDocs: ArgsDocs, reqdArgs: Seq[str] = None):
        self.argsDocs = allArgsDocs
        self.reqdArgs = reqdArgs or ()
        self.argsSpecType = ArgSpecGiven.SPEC_GIVEN_ALL

    def isComplete(self) -> bool:
        """Returns True if the definition is usable, false otherwise."""
        return self.argsSpecType == ArgSpecGiven.SPEC_GIVEN_ALL

    def getOptional(self) -> List[str]:
        """Get the list of optional arguments"""
        return tuple(set(self.argsDocs.keys()).difference(self.reqdArgs))

    def __str__(self):
        return "%s, %s, %s" % (self.argsDocs, self.reqdArgs, self.argsSpecType)


class SenderMissingReqdMsgDataError(RuntimeError):
    """
    Raised when a sendMessage() is missing arguments tagged as
    'required' by pubsub topic of message.
    """

    def __init__(self, topicName: str, argNames: Seq[str], missing: Seq[str]):
        argsStr = ','.join(argNames)
        missStr = ','.join(missing)
        msg = "Some required args missing in call to sendMessage('%s', %s): %s" \
              % (stringize(topicName), argsStr, missStr)
        RuntimeError.__init__(self, msg)


class SenderUnknownMsgDataError(RuntimeError):
    """
    Raised when a sendMessage() has arguments not listed among the topic's
    message data specification (MDS).
    """

    def __init__(self, topicName: str, argNames: Seq[str], extra: Seq[str]):
        argsStr = ','.join(argNames)
        extraStr = ','.join(extra)
        msg = "Some optional args unknown in call to sendMessage('%s', %s): %s" \
              % (topicName, argsStr, extraStr)
        RuntimeError.__init__(self, msg)


@annotationType
class ArgsInfo:
    pass


class ArgsInfo:
    """
    Encode the Message Data Specification (MDS) for a given
    topic. ArgsInfos form a tree identical to that of Topics in that
    ArgInfos have a reference to their parent and children ArgInfos,
    created for the parent and children topics.

    The only difference
    between an ArgsInfo and an ArgSpecGiven is that the latter is
    what "user thinks is ok" whereas former has been validated:
    the specification for this topic is a strict superset of the
    specification of its parent, and a strict subset of the
    specification of each of its children. Also, the instance
    can be used to check validity and filter arguments.

    The MDS can be created "empty", ie "incomplete", meaning it cannot
    yet be used to validate listener subscriptions to topics.
    """

    SPEC_MISSING = 10  # no args given
    SPEC_COMPLETE = 12  # all args, but not confirmed via user spec

    def __init__(self, topicNameTuple: Seq[str], specGiven: ArgSpecGiven, parentArgsInfo: ArgsInfo):
        self.topicNameTuple = topicNameTuple
        self.allOptional = ()  # topic message optional arg names
        self.allDocs = {}  # doc for each arg
        self.allRequired = ()  # topic message required arg names
        self.argsSpecType = self.SPEC_MISSING
        self.parentAI = WeakNone()
        if parentArgsInfo is not None:
            self.parentAI = weakref.ref(parentArgsInfo)
            parentArgsInfo.__addChildAI(self)
        self.childrenAI = []

        if specGiven.isComplete():
            self.__setAllArgs(specGiven)

        if parentArgsInfo is None:
            assert self.argsAddedToParent is not None
        else:
            while not parentArgsInfo.isComplete():
                parentArgsInfo = parentArgsInfo.parentAI()
            self.argsAddedToParent = set(self.getArgs()).difference(parentArgsInfo.getArgs())


    def isComplete(self) -> bool:
        return self.argsSpecType == self.SPEC_COMPLETE

    def getArgs(self) -> List[str]:
        return self.allOptional + self.allRequired

    def numArgs(self) -> int:
        return len(self.allOptional) + len(self.allRequired)

    def getReqdArgs(self) -> List[str]:
        return self.allRequired

    def getOptArgs(self) -> List[str]:
        return self.allOptional

    def getArgsDocs(self) -> ArgsDocs:
        return self.allDocs.copy()

    def setArgsDocs(self, docs: ArgsDocs):
        """docs is a mapping from arg names to their documentation"""
        if not self.isComplete():
            raise RuntimeError('Topic MDS is not complete, cannot set docs!')
        for arg, doc in docs.items():
            self.allDocs[arg] = doc

    def check(self, msgData: MsgData):
        """
        Check that the message arguments given satisfy the topic message
        data specification (MDS).
        :param msgData: the topic message data to check for validity
        :raise SenderMissingReqdMsgDataError: if some required args are missing or not known
        :raise SenderUnknownMsgDataError: if some optional args are unknown.
        """
        all = set(msgData)
        # check that it has all required args
        needReqd = set(self.allRequired)
        hasReqd = (needReqd <= all)
        if not hasReqd:
            raise SenderMissingReqdMsgDataError(
                self.topicNameTuple, list(msgData.keys()), needReqd - all)

        # check that all other args are among the optional spec
        optional = all - needReqd
        ok = (optional <= set(self.allOptional))
        if not ok:
            raise SenderUnknownMsgDataError(self.topicNameTuple,
                                            list(msgData.keys()), optional - set(self.allOptional))

    def filterArgs(self, msgData: MsgData) -> MsgData:
        """
        Returns a dict which contains only those items of msgData which are defined for topic.
        E.g. if msgData is {a:1, b:'b'} and topic arg spec is ('a',) then return {a:1}. The returned
        dict is valid only if check(msgData) was called (or check(superset of msgData) was called).

        :param msgData: the topic message data to filter
        """
        assert self.isComplete()
        if len(msgData) == self.numArgs():
            return msgData

        # only keep the keys from msgData that are also in topic's kwargs
        # method 1: SLOWEST
        # newKwargs = dict( (k,msgData[k]) for k in self.__msgArgs.allOptional if k in msgData )
        # newKwargs.update( (k,msgData[k]) for k in self.__msgArgs.allRequired )

        # method 2: FAST:
        # argNames = self.__msgArgs.getArgs()
        # newKwargs = dict( (key, val) for (key, val) in msgData.iteritems() if key in argNames )

        # method 3: FASTEST:
        argNames = set(self.getArgs()).intersection(msgData)
        newKwargs = dict((k, msgData[k]) for k in argNames)

        return newKwargs

    def hasSameArgs(self, *argNames: Seq[str]) -> bool:
        """
        Returns true if self has all the message arguments given, no
        more and no less. Order does not matter. So if getArgs()
        returns ('arg1', 'arg2') then self.hasSameArgs('arg2', 'arg1')
        will return true.
        """
        return set(argNames) == set(self.getArgs())

    def hasParent(self, argsInfo: ArgsInfo) -> bool:
        """return True if self has argsInfo object as parent"""
        return self.parentAI() is argsInfo

    def getCompleteAI(self) -> ArgsInfo:
        """
        Get the closest arg spec, starting from self and moving to parent,
        that is complete. So if self.isComplete() is True, then returns self,
        otherwise returns parent (if parent.isComplete()), etc.
        """
        AI = self
        while AI is not None:
            if AI.isComplete():
                return AI
            AI = AI.parentAI()  # dereference weakref
        return None

    def updateAllArgsFinal(self, topicDefn: ArgSpecGiven):
        """
        This can only be called once, if the construction was done
        with ArgSpecGiven.SPEC_GIVEN_NONE
        """
        assert not self.isComplete()
        assert topicDefn.isComplete()
        self.__setAllArgs(topicDefn)

    def __addChildAI(self, childAI: ArgsInfo):
        assert childAI not in self.childrenAI
        self.childrenAI.append(childAI)

    def __notifyParentCompleted(self):
        """Parent should call this when parent ArgsInfo has been completed"""
        assert self.parentAI().isComplete()
        if self.isComplete():
            # verify that our spec is compatible with parent's
            self.__validateArgsToParent()
            self.argsAddedToParent = set(self.getArgs()).difference(self.parentAI().getArgs())
        else:
            for argsInfo in self.childrenAI:
                argsInfo.__notifyAncestorCompleted(self.parentAI())

    def __notifyAncestorCompleted(self, parentAI):
        if self.isComplete():
            # verify that our spec is compatible with parent's
            self.__validateArgsToParent()
            self.argsAddedToParent = set(self.getArgs()).difference(parentAI.getArgs())
        else:
            for argsInfo in self.childrenAI:
                argsInfo.__notifyAncestorCompleted(parentAI)

    def __validateArgsToParent(self):
        # validate relative to parent arg spec
        closestParentAI = self.parentAI().getCompleteAI()
        if closestParentAI is not None:
            # verify that parent args is a subset of spec given:
            topicName = stringize(self.topicNameTuple)
            verifySubset(self.getArgs(), closestParentAI.getArgs(), topicName)
            verifySubset(self.allRequired, closestParentAI.getReqdArgs(), topicName, ' required args')

    def __setAllArgs(self, specGiven: ArgSpecGiven):
        assert specGiven.isComplete()
        self.allOptional = tuple(specGiven.getOptional())
        self.allRequired = specGiven.reqdArgs
        self.allDocs = specGiven.argsDocs.copy()  # doc for each arg
        self.argsSpecType = self.SPEC_COMPLETE

        parentArgsInfo = self.parentAI()
        if parentArgsInfo is None:
            self.argsAddedToParent = []
        else:
            self.__validateArgsToParent()
            while not parentArgsInfo.isComplete():
                parentArgsInfo = parentArgsInfo.parentAI()
            self.argsAddedToParent = set(self.getArgs()).difference(parentArgsInfo.getArgs())

        # notify our children
        for childAI in self.childrenAI:
            childAI.__notifyParentCompleted()



"""
Code related to the concept of topic tree and its management: creating
and removing topics, getting info about a particular topic, etc.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from typing import Tuple, List, Sequence, Mapping, Dict, Callable, Any, Optional, Union, TextIO

from .callables import getID, UserListener

from .topicutils import (
    ALL_TOPICS,
    tupleize,
    stringize,
)

from .topicexc import (
    TopicNameError,
    TopicDefnError,
)

from .topicargspec import (
    ArgSpecGiven,
    topicArgsFromCallable,
    ArgsInfo,
)

from .topicobj import Topic
from .listener import IListenerExcHandler
from .topicdefnprovider import ITopicDefnProvider
from .notificationmgr import NotificationMgr, INotificationHandler

# ---------------------------------------------------------

__all__ = [
    'TopicManager',
    'TopicNameError',
    'TopicDefnError',
]

ARGS_SPEC_ALL = ArgSpecGiven.SPEC_GIVEN_ALL
ARGS_SPEC_NONE = ArgSpecGiven.SPEC_GIVEN_NONE


# ---------------------------------------------------------

class TreeConfig:
    """
    Each topic tree has its own topic manager and configuration,
    such as notification and exception handling.
    """

    def __init__(self, notificationHandler: INotificationHandler = None,
                 listenerExcHandler: IListenerExcHandler = None):
        self.notificationMgr = NotificationMgr(notificationHandler)
        self.listenerExcHandler = listenerExcHandler
        self.raiseOnTopicUnspecified = False


class TopicManager:
    """
    Manages the registry of all topics and creation/deletion
    of topics.

    Note that any method that accepts a topic name can accept it in the
    'dotted' format such as ``'a.b.c.'`` or in tuple format such as
    ``('a', 'b', 'c')``. Any such method will raise a ValueError
    if name not valid (empty, invalid characters, etc).
    """

    # Allowed return values for isTopicSpecified()
    TOPIC_SPEC_NOT_SPECIFIED = 0  # false
    TOPIC_SPEC_ALREADY_CREATED = 1  # all other values equate to "true" but different reason
    TOPIC_SPEC_ALREADY_DEFINED = 2

    def __init__(self, treeConfig: TreeConfig = None):
        """
        The optional treeConfig is an instance of TreeConfig, used to
        configure the topic tree such as notification settings, etc. A
        default config is created if not given. This method should only be
        called by an instance of Publisher (see Publisher.getTopicManager()).
        """
        self.__allTopics = None  # root of topic tree
        self._topicsMap = {}  # registry of all topics
        self.__treeConfig = treeConfig or TreeConfig()
        self.__defnProvider = _MasterTopicDefnProvider(self.__treeConfig)

        # define root of all topics
        assert self.__allTopics is None
        argsDocs, reqdArgs = {}, ()
        desc = 'Root of all topics'
        specGiven = ArgSpecGiven(argsDocs, reqdArgs)
        self.__allTopics = self.__createTopic((ALL_TOPICS,), desc, specGiven=specGiven)

    def getRootAllTopics(self) -> Topic:
        """
        Get the topic that is parent of all root (ie top-level) topics,
        for default TopicManager instance created when this module is imported.
        Some notes:

        - "root of all topics" topic satisfies isAll()==True, isRoot()==False,
          getParent() is None;
        - all root-level topics satisfy isAll()==False, isRoot()==True, and
          getParent() is getDefaultTopicTreeRoot();
        - all other topics satisfy neither.
        """
        return self.__allTopics

    def addDefnProvider(self, providerOrSource: Any, format=None) -> ITopicDefnProvider:
        """
        Register a topic definition provider. After this method is called, whenever a topic must
        be created, the first definition provider that has a definition for the required topic
        is used to instantiate the topic.

        If providerOrSource is an instance of ITopicDefnProvider, register
        it as a provider of topic definitions. Otherwise, register a new
        instance of TopicDefnProvider(providerOrSource, format). In that case,
        if format is not given, it defaults to TOPIC_TREE_FROM_MODULE. Either
        way, returns the instance of ITopicDefnProvider registered.
        """
        if isinstance(providerOrSource, ITopicDefnProvider):
            provider = providerOrSource
        else:
            from .topicdefnprovider import TopicDefnProvider, TOPIC_TREE_FROM_MODULE
            source = providerOrSource
            provider = TopicDefnProvider(source, format or TOPIC_TREE_FROM_MODULE)
        self.__defnProvider.addProvider(provider)
        return provider

    def clearDefnProviders(self):
        """Remove all registered topic definition providers"""
        self.__defnProvider.clear()

    def getNumDefnProviders(self) -> int:
        """Get how many topic definitions providers are registered."""
        return self.__defnProvider.getNumProviders()

    def getTopic(self, name: str, okIfNone: bool = False) -> Topic:
        """
        Get the Topic instance for the given topic name. By default, raises
        an TopicNameError exception if a topic with given name doesn't exist. If
        okIfNone=True, returns None instead of raising an exception.
        """
        topicNameDotted = stringize(name)
        # if not name:
        #    raise TopicNameError(name, 'Empty topic name not allowed')
        obj = self._topicsMap.get(topicNameDotted, None)
        if obj is not None:
            return obj

        if okIfNone:
            return None

        # NOT FOUND! Determine what problem is and raise accordingly:
        # find the closest parent up chain that does exists:
        parentObj, subtopicNames = self.__getClosestParent(topicNameDotted)
        assert subtopicNames

        subtopicName = subtopicNames[0]
        if parentObj is self.__allTopics:
            raise TopicNameError(name, 'Root topic "%s" doesn\'t exist' % subtopicName)

        msg = 'Topic "%s" doesn\'t have "%s" as subtopic' % (parentObj.getName(), subtopicName)
        raise TopicNameError(name, msg)

    def getOrCreateTopic(self, name: str, protoListener: UserListener = None) -> Topic:
        """
        Get the Topic instance for topic of given name, creating it
        (and any of its missing parent topics) as necessary. Pubsub
        functions such as subscribe() use this to obtain the Topic object
        corresponding to a topic name.

        The name can be in dotted or string format (``'a.b.'`` or ``('a','b')``).

        This method always attempts to return a "complete" topic, i.e. one
        with a Message Data Specification (MDS). So if the topic does not have
        an MDS, it attempts to add it. It first tries to find an MDS
        from a TopicDefnProvider (see addDefnProvider()). If none is available,
        it attempts to set it from protoListener, if it has been given. If not,
        the topic has no MDS.

        Once a topic's MDS has been set, it is never again changed or accessed
        by this method.

        Examples::

            # assume no topics exist
            # but a topic definition provider has been added via
            # pub.addTopicDefnProvider() and has definition for topics 'a' and 'a.b'

            # creates topic a and a.b; both will have MDS from the defn provider:
            t1 = topicMgr.getOrCreateTopic('a.b')
            t2 = topicMgr.getOrCreateTopic('a.b')
            assert(t1 is t2)
            assert(t1.getParent().getName() == 'a')

            def proto(req1, optarg1=None): pass
            # creates topic c.d with MDS based on proto; creates c without an MDS
            # since no proto for it, nor defn provider:
            t1 = topicMgr.getOrCreateTopic('c.d', proto)

        The MDS can also be defined via a call to subscribe(listener, topicName),
        which indirectly calls getOrCreateTopic(topicName, listener).
        """
        obj = self.getTopic(name, okIfNone=True)
        if obj:
            # if object is not sendable but a proto listener was given,
            # update its specification so that it is sendable
            if (protoListener is not None) and not obj.hasMDS():
                allArgsDocs, required = topicArgsFromCallable(protoListener)
                obj.setMsgArgSpec(allArgsDocs, required)
            return obj

        # create missing parents
        nameTuple = tupleize(name)
        parentObj = self.__createParentTopics(nameTuple)

        # now the final topic object, args from listener if provided
        desc, specGiven = self.__defnProvider.getDefn(nameTuple)
        # POLICY: protoListener is used only if no definition available
        if specGiven is None:
            if protoListener is None:
                desc = 'UNDOCUMENTED: created without spec'
            else:
                allArgsDocs, required = topicArgsFromCallable(protoListener)
                specGiven = ArgSpecGiven(allArgsDocs, required)
                desc = 'UNDOCUMENTED: created from protoListener "%s" in module %s' % getID(protoListener)

        return self.__createTopic(nameTuple, desc, parent=parentObj, specGiven=specGiven)

    def isTopicInUse(self, name: str) -> bool:
        """
        Determine if topic 'name' is in use. True if a Topic object exists
        for topic name (i.e. message has already been sent for that topic, or a
        least one listener subscribed), false otherwise. Note: a topic may be in use
        but not have a definition (MDS and docstring); or a topic may have a
        definition, but not be in use.
        """
        return self.getTopic(name, okIfNone=True) is not None

    def hasTopicDefinition(self, name: str) -> bool:
        """
        Determine if there is a definition avaiable for topic 'name'. Return
        true if there is, false otherwise. Note: a topic may have a
        definition without being in use, and vice versa.
        """
        # in already existing Topic object:
        alreadyCreated = self.getTopic(name, okIfNone=True)
        if alreadyCreated is not None and alreadyCreated.hasMDS():
            return True

        # from provider?
        nameTuple = tupleize(name)
        if self.__defnProvider.isDefined(nameTuple):
            return True

        return False

    def checkAllTopicsHaveMDS(self):
        """
        Check that all topics that have been created for their MDS.
        Raise a TopicDefnError if one is found that does not have one.
        """
        for topic in self._topicsMap.values():
            if not topic.hasMDS():
                raise TopicDefnError(topic.getNameTuple())

    def delTopic(self, name: str) -> bool:
        """
        Delete the named topic, including all sub-topics. Returns False
        if topic does not exist; True otherwise. Also unsubscribe any listeners
        of topic and all subtopics.
        """
        # find from which parent the topic object should be removed
        dottedName = stringize(name)
        try:
            # obj = weakref( self._topicsMap[dottedName] )
            obj = self._topicsMap[dottedName]
        except KeyError:
            return False

        # assert obj().getName() == dottedName
        assert obj.getName() == dottedName
        # notification must be before deletion in case
        self.__treeConfig.notificationMgr.notifyDelTopic(dottedName)

        # obj()._undefineSelf_(self._topicsMap)
        obj._undefineSelf_(self._topicsMap)
        # assert obj() is None

        return True

    def getTopicsSubscribed(self, listener: UserListener) -> List[Topic]:
        """
        Get the list of Topic objects that have given listener
        subscribed. Note: the listener can also get messages from any
        sub-topic of returned list.
        """
        assocTopics = []
        for topicObj in self._topicsMap.values():
            if topicObj.hasListener(listener):
                assocTopics.append(topicObj)
        return assocTopics

    def clearTree(self):
        """Remove every topic from the topic tree"""
        for topic in list(self.__allTopics.subtopics):
            self.delTopic(topic.name)

    def __getClosestParent(self, topicNameDotted: str) -> Topic:
        """
        Returns a pair, (closest parent, tuple path from parent). The
        first item is the closest parent Topic that exists.
        The second one is the list of topic name elements that have to be
        created to create the given topic.

        So if topicNameDotted = A.B.C.D, but only A.B exists (A.B.C and
        A.B.C.D not created yet), then return is (A.B, ['C','D']).
        Note that if none of the branch exists (not even A), then return
        will be [root topic, ['A',B','C','D']). Note also that if A.B.C
        exists, the return will be (A.B.C, ['D']) regardless of whether
        A.B.C.D exists.
        """
        subtopicNames = []
        headTail = topicNameDotted.rsplit('.', 1)
        while len(headTail) > 1:
            parentName = headTail[0]
            subtopicNames.insert(0, headTail[1])
            obj = self._topicsMap.get(parentName, None)
            if obj is not None:
                return obj, subtopicNames

            headTail = parentName.rsplit('.', 1)

        subtopicNames.insert(0, headTail[0])
        return self.__allTopics, subtopicNames

    def __createParentTopics(self, topicName: str) -> Topic:
        """
        This will find which parents need to be created such that
        topicName can be created (but doesn't create given topic),
        and creates them. Returns the parent object.
        """
        assert self.getTopic(topicName, okIfNone=True) is None
        parentObj, subtopicNames = self.__getClosestParent(stringize(topicName))

        # will create subtopics of parentObj one by one from subtopicNames
        if parentObj is self.__allTopics:
            nextTopicNameList = []
        else:
            nextTopicNameList = list(parentObj.getNameTuple())
        for name in subtopicNames[:-1]:
            nextTopicNameList.append(name)
            desc, specGiven = self.__defnProvider.getDefn(tuple(nextTopicNameList))
            if desc is None:
                desc = 'UNDOCUMENTED: created as parent without specification'
            parentObj = self.__createTopic(tuple(nextTopicNameList),
                                           desc, specGiven=specGiven, parent=parentObj)

        return parentObj

    def __createTopic(self, nameTuple: Sequence[str], desc: str, specGiven: ArgSpecGiven,
                      parent: Topic = None) -> Topic:
        """
        Actual topic creation step. Adds new Topic instance to topic map,
        and sends notification message (see ``Publisher.addNotificationMgr()``)
        regarding topic creation.
        """
        if specGiven is None:
            specGiven = ArgSpecGiven()
        parentAI = None
        if parent:
            parentAI = parent._getListenerSpec()
        argsInfo = ArgsInfo(nameTuple, specGiven, parentAI)
        if (self.__treeConfig.raiseOnTopicUnspecified
            and not argsInfo.isComplete()):
            raise TopicDefnError(nameTuple)

        newTopicObj = Topic(self.__treeConfig, nameTuple, desc,
                            argsInfo, parent=parent)
        # sanity checks:
        assert newTopicObj.getName() not in self._topicsMap
        if parent is self.__allTopics:
            assert len(newTopicObj.getNameTuple()) == 1
        else:
            assert parent.getNameTuple() == newTopicObj.getNameTuple()[:-1]
        assert nameTuple == newTopicObj.getNameTuple()

        # store new object and notify of creation
        self._topicsMap[newTopicObj.getName()] = newTopicObj
        self.__treeConfig.notificationMgr.notifyNewTopic(
            newTopicObj, desc, specGiven.reqdArgs, specGiven.argsDocs)

        return newTopicObj


def validateNameHierarchy(topicTuple: Tuple[Topic, ...]):
    """
    Check that names in topicTuple are valid: no spaces, not empty.
    Raise ValueError if fails check. E.g. ('',) and ('a',' ') would
    both fail, but ('a','b') would be ok.
    """
    if not topicTuple:
        topicName = stringize(topicTuple)
        errMsg = 'empty topic name'
        raise TopicNameError(topicName, errMsg)

    for indx, topic in enumerate(topicTuple):
        errMsg = None
        if topic is None:
            topicName = list(topicTuple)
            topicName[indx] = 'None'
            errMsg = 'None at level #%s'

        elif not topic:
            topicName = stringize(topicTuple)
            errMsg = 'empty element at level #%s'

        elif topic.isspace():
            topicName = stringize(topicTuple)
            errMsg = 'blank element at level #%s'

        if errMsg:
            raise TopicNameError(topicName, errMsg % indx)


class _MasterTopicDefnProvider:
    """
    Stores a list of topic definition providers. When queried for a topic
    definition, queries each provider (registered via addProvider()) and
    returns the first complete definition provided, or (None,None).

    The providers must follow the ITopicDefnProvider API.
    """

    def __init__(self, treeConfig: TreeConfig):
        self.__providers = []
        self.__treeConfig = treeConfig

    def addProvider(self, provider):
        """Add given provider IF not already added. """
        assert (isinstance(provider, ITopicDefnProvider))
        if provider not in self.__providers:
            self.__providers.append(provider)

    def clear(self):
        """Remove all providers added."""
        self.__providers = []

    def getNumProviders(self) -> int:
        """Return how many providers added."""
        return len(self.__providers)

    def getDefn(self, topicNameTuple: Sequence[str]) -> Tuple[str, ArgSpecGiven]:
        """
        Returns a pair (docstring, MDS) for the topic. The first item is
        a string containing the topic's "docstring", i.e. a description string
        for the topic, or None if no docstring available for the topic. The
        second item is None or an instance of ArgSpecGiven specifying the
        required and optional message data for listeners of this topic.
        """
        desc, defn = None, None
        for provider in self.__providers:
            tmpDesc, tmpDefn = provider.getDefn(topicNameTuple)
            if (tmpDesc is not None) and (tmpDefn is not None):
                assert tmpDefn.isComplete()
                desc, defn = tmpDesc, tmpDefn
                break

        return desc, defn

    def isDefined(self, topicNameTuple: Sequence[str]) -> bool:
        """
        Returns True only if a complete definition exists, ie topic
        has a description and a complete message data specification (MDS).
        """
        desc, defn = self.getDefn(topicNameTuple)
        if desc is None or defn is None:
            return False
        if defn.isComplete():
            return True
        return False



"""
Provides useful functions and classes. Most useful are probably
printTreeDocs and printTreeSpec.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

import sys

__all__ = ('printImported', 'Callback')


def printImported():
    """Output a list of pubsub modules imported so far"""
    ll = [mod for mod in sys.modules.keys() if mod.find('pubsub') >= 0]  # iter keys ok
    ll.sort()
    print('\n'.join(ll))


class Callback:
    """
    This can be used to wrap functions that are referenced by class
    data if the data should be called as a function. E.g. given
    >>> def func(): pass
    >>> class A:
    ....def __init__(self): self.a = func
    then doing
    >>> boo=A(); boo.a()
    will fail since Python will try to call a() as a method of boo,
    whereas a() is a free function. But if you have instead
    "self.a = Callback(func)", then "boo.a()" works as expected.
    """

    def __init__(self, callable_obj):
        self.__callable = callable_obj

    def __call__(self, *args, **kwargs):
        return self.__callable(*args, **kwargs)



"""
Low level functions and classes related to callables.

The AUTO_TOPIC
is the "marker" to use in callables to indicate that when a message
is sent to those callables, the topic object for that message should be
added to the data sent via the call arguments. See the docs in
CallArgsInfo regarding its autoTopicArgName data member.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.

"""

from inspect import ismethod, isfunction, signature, Parameter
import sys
from types import ModuleType
from typing import Tuple, List, Sequence, Callable, Any

# Opaque constant used to mark a kwarg of a listener as one to which pubsub should assign the topic of the
# message being sent to the listener. This constant should be used by reference; its value is "unique" such that
# pubsub can find such kwarg.
class AUTO_TOPIC: pass

# In the user domain, a listener is any callable, regardless of signature. The return value is ignored,
# i.e. the listener will be treated as though it is a Callable[..., None]. Also, the args, "...", must be
# consistent with the MDS of the topic to which listener is being subscribed.
UserListener = Callable[..., Any]


def getModule(obj: Any) -> ModuleType:
    """
    Get the module in which an object was defined.
    :param obj: the object for which to get module
    :return: the module object, or the string '__main__' if no module defined for obj
        (which usually indicates either a builtin, or a definition within main script).
    """
    if hasattr(obj, '__module__'):
        module = obj.__module__
    else:
        module = '__main__'
    return module


def getID(callable_obj: UserListener) -> Tuple[str, ModuleType]:
    """
    Get "ID" of a callable, in the form of its name and module in which it is defined
    E.g. getID(Foo.bar) returns ('Foo.bar', 'a.b') if Foo.bar was defined in module a.b.
    :param callable_obj: a callable, ie function, bound method or callable instance
    """
    sc = callable_obj
    if ismethod(sc):
        module = getModule(sc.__self__)
        obj_name = '%s.%s' % (sc.__self__.__class__.__name__, sc.__func__.__name__)
    elif isfunction(sc):
        module = getModule(sc)
        obj_name = sc.__name__
    else:  # must be a functor (instance of a class that has __call__ method)
        module = getModule(sc)
        obj_name = sc.__class__.__name__

    return obj_name, module


def getRawFunction(callable_obj: UserListener) -> Tuple[Callable]:
    """
    Get raw function information about a callable.
    :param callable_obj: any object that can be called
    :return: function corresponding to callable, and offset is 0 or 1 to
        indicate whether the function's first argument is 'self' (1) or not (0)
    :raise ValueError: if callable_obj is not of a recognized type (function, method or object with __call__ method).
    """
    firstArg = 0
    if isfunction(callable_obj):
        # print 'Function', getID(callable_obj)
        func = callable_obj
    elif ismethod(callable_obj):
        # print 'Method', getID(callable_obj)
        func = callable_obj
    elif hasattr(callable_obj, '__call__'):
        # print 'Functor', getID(callable_obj)
        func = callable_obj.__call__
    else:
        msg = 'type "%s" not supported' % type(callable_obj).__name__
        raise ValueError(msg)

    return func


class ListenerMismatchError(ValueError):
    """
    Raised when an attempt is made to subscribe a listener to
    a topic, but listener does not satisfy the topic's message data
    specification (MDS). This specification is inferred from the first
    listener subscribed to a topic, or from an imported topic tree
    specification (see pub.addTopicDefnProvider()).
    """

    def __init__(self, msg: str, listener: UserListener, *args):
        idStr, module = getID(listener)
        msg = 'Listener "%s" (from module "%s") inadequate: %s' % (idStr, module, msg)
        ValueError.__init__(self, msg)
        self.args = args
        self.msg = msg
        self.module = module
        self.idStr = idStr

    def __str__(self):
        return self.msg


class CallArgsInfo:
    """
    Represent the "signature" of a listener of topic messages: which arguments are
    required vs optional.
    """

    def __init__(self, func: UserListener, ignoreArgs: Sequence[str] = ()):
        """
        :param func: the callable for which to get paramaters info
        :param ignoreArgs: do not include the given names in the get*Args() return values

        After construction,
        - self.acceptsAllKwargs = True if the listener has a **kwargs arg
        - self.autoTopicArgName will be the name of argument in which to put the Topic
          object for which pubsub message is sent, or None if auto off. This is identified
          by a parameter that has a default value of AUTO_TOPIC.

        For instance,
        - listener(self, arg1, arg2=AUTO_TOPIC, arg3=None) will have self.allParams = (arg1, arg2, arg3),
            self.numRequired=1, and self.autoTopicArgName = 'arg2', whereas
        - listener(self, arg1, arg3=None) will have self.allParams = (arg1, arg3), self.numRequired=1, and
            self.autoTopicArgName = None.
        """

        requiredArgs = []
        optionalArgs = []
        self.autoTopicArgName = None
        self.acceptsAllKwargs = False
        for argName, param in signature(func).parameters.items():
            if argName in ignoreArgs or param.kind == Parameter.VAR_POSITIONAL:
                continue

            if param.kind == Parameter.VAR_KEYWORD:
                self.acceptsAllKwargs = True
                continue

            if param.default == Parameter.empty:
                requiredArgs.append(argName)
            else:
                if param.default == AUTO_TOPIC:
                    self.autoTopicArgName = argName
                else:
                    optionalArgs.append(argName)

        self.requiredArgs = tuple(requiredArgs)
        self.optionalArgs = tuple(optionalArgs)
        self.allParams = self.requiredArgs + self.optionalArgs

    def getAllArgs(self) -> Tuple[str]:
        """
        Return a tuple of names indicating the complete set of message data
        (keyword args) that can be given to this listener
        """
        return self.optionalArgs

    def getOptionalArgs(self) -> Tuple[str]:
        """
        Return a tuple of names indicating which message data (keyword args)
        are optional when this listener is called.
        """
        return self.optionalArgs

    def getRequiredArgs(self) -> Tuple[str]:
        """
        Return a tuple of names indicating which message data (keyword args)
        are required when this listener is called.
        """
        return self.requiredArgs


def getArgs(callable_obj: UserListener, ignoreArgs: Sequence[str] = ()) -> CallArgsInfo:
    """
    Get the call parameters of a callable to be used as listener.
    :param callable_obj: the callable for which to get call parameters
    :param ignoreArgs: optional list of names of parameters of callable_obj that should not be in the returned object
    :return: an instance of CallArgsInfo for the given callable_obj
    :raise ListenerMismatchError: if callable_obj is not a callable, or ignoreArgs has an item that is not a call
        param of callable
    """
    # figure out what is the actual function object to inspect:
    try:
        func = getRawFunction(callable_obj)
    except ValueError:
        exc = sys.exc_info()[1]
        raise ListenerMismatchError(str(exc), callable_obj)

    return CallArgsInfo(func, ignoreArgs=ignoreArgs)
